/*
		"@(#) autologout.c by David Dickson" 
               
                Modified by John Auckett 16/09/94

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/
#include    <stdio.h>
#include    <signal.h>
#include    <string.h>
#include    <sys/types.h>
#include    <sys/stat.h>
#include    <utmp.h>
#define     GRACE       60  /* grace time (sec) for user reply */
#define     KWAIT       20  /* time to wait after kill (in sec.) */
#define     WARNING     1   /* a warning message */
#define     LOGOFF      2   /* a log-off message */
#define     NOLOGOFF    3   /* a log-off failure message */
#define     SEC         60  /* seconds per minute */
#define     TIMEMAX     10  /* maximum idle time (minutes) */

#define     NO_OF_LINES 6   /* number of dial-in lines installed */
#define     REPRIEVE    4   /* line became available during grace time */
#define     EXTRATIME   5   /* go into borrowed time message */
#define     DEFAULTTIME 10  /* default time limit */

struct utmp *utmpp;         /* pointer to utmp file entry */
char        *ctime();       /* returns pointer to time string */
struct utmp *getutent();    /* returns next utmp file entry */
void         setutent();
FILE        *ufp;

int free_lines()    /* returns the number of free dial-in lines */
{
  int line_count = 0;
  struct utmp *upp;
  
  /* count up number of ttyS devices */
  setutent();
  while ((upp = getutent()) != (struct utmp *) NULL)
    if ((upp->ut_type == USER_PROCESS) && (upp->ut_line[3] == 'S'))
      ++line_count;
/*  printf("Lines in use %d\n", line_count); */
  return(NO_OF_LINES - line_count);
}

int time_permitted(char *name)  /* returns a users time per call */
{
  FILE *fp;
  char record_name[9];
  int time_per_call;

  /* fetch a user record from dialin.users */
  if (!(fp = fopen("/usr/etc/dialin.users", "r")))
  {                   /* if the file can't be opened then    */
    fclose(fp);       /* use a default time limit            */
    return(DEFAULTTIME);
  }
  while(1)     /* continue until name found or can't read */
  {            /* anymore names from file                 */
    if (fscanf(fp, "%s %d", record_name, &time_per_call) != 2)
    {                          /* got to end without finding name */
      fclose(fp);              /* name not in the dialin.users    */
      return(DEFAULTTIME);     /* file so use default time limit  */
    }
    if (!strcmp(name, record_name))
    {   /* found the users record in the file */
      fclose(fp);
      return(time_per_call);
    }
  }
}

main()
{ 
  int count=0;
  
    while ((utmpp = getutent()) != (struct utmp *) NULL)
    {
      if (utmpp->ut_type == USER_PROCESS)
        if (fork())     /* parent process continues thru utmp entries */
          ;
        else
        {
          check_idle(); /* child checks utmp entries */
        }
     }
}

check_idle()    /* select utmp entries older than TIMEMAX */
{
    char        dev[24], name[12];
    struct stat status, *pstat;
    time_t      online, pres_time, start, time();
    int         time_allowed;
    int         user_pid;
    struct utmp *newutmp;
    uid_t       user_id;

    pstat = &status;    /* point to status structure */

   if (utmpp->ut_line[3] != 'S')
   /*  && (utmpp->ut_line[3] != 'p'))  */
                 /* if not a dial in line or a telnet*/
        exit(0);                        /* skip the utmp entry   */

    strcpy(dev, "/dev/");           /* add "/dev/" directory prefix */
    strcat(dev, utmpp->ut_line);    /* append basename of port */
    if (stat(dev, pstat))           /* if can't get status for port */
        bailout("can't get status of user's terminal", 1);
    pres_time = time((time_t *)0);  /* get current time */

    online = (pres_time - utmpp->ut_time) /SEC;
 
    strncpy(name, utmpp->ut_user, 8);   /* else get user name */
    name[8] = '\0';          /* null terminate user name string */
    
    /* calling free_lines() corrupts the utmpp pointer */ 
    user_pid = utmpp->ut_pid;

    time_allowed = time_permitted(name);
    if (online < time_allowed)
      exit(0);

/*
    printf("%s  %s has been online for %d mins\n", ctime(&pres_time), name, online);
    printf("    %d free lines available\n", free_lines());
    printf("    time permitted is %d mins\n", time_allowed);
    printf("pid : %d\n", user_pid);
*/

    mesg(WARNING, name, dev, (int)online); /* send warning to user */
    if (stat(dev, pstat))
        bailout("can't get status of user's terminal", 2);
    start = pstat->st_atime;    /* start time for countdown */
    
    if (time_allowed > 25)    /* only allow borrowed time for */
      if (free_lines() > 0)   /* non-guest accounts           */
      {
          mesg(EXTRATIME, name, dev, (int)online);
          exit(0);
      }

user_id = pstat->st_uid;  /* remember the user id */

/* sleep for 2 mins */
    sleep(GRACE*2);
    
    if (stat(dev, pstat))
        bailout("can't get status of user's terminal", 3);
    if (pstat->st_uid != user_id)
        bailout("autologout left over from dropped carrier", 3);

    if (time_allowed > 25)    /* only all borowed time for */
      if (free_lines() > 0)     /* non-guests                */
      {
          mesg(REPRIEVE, name, dev, (int)online);
          exit(0);
      }
      /* user has run out of time */
    if (!killit(user_pid))
      mesg(NOLOGOFF, name, dev, (int)online); /* couldn't kill */
    else
      mesg(LOGOFF, name, dev, (int)online); /* okay */
    exit(0);
}

mesg(flag, name, dev, idle) /* mail to user and log message */
int     flag, idle;     /* flag indicates message type */
char    *name, *dev;
{
    char    mbuf[256];  /* message buffer */
    time_t  tvec;
    FILE    *fopen(), *fp, *log, *mprog;

    time(&tvec);    /* store time in tvec */
    if ((log = fopen("/usr/adm/logoutlog", "a")) == (FILE *) NULL)
        bailout("can't open log file", 4);

    if (flag == EXTRATIME)
    {
      if ((fp = fopen(dev, "w")) == (FILE *) NULL)
          bailout("can't open user's terminal", 5);
      fprintf(fp, "But since a line is free you can stay...\n\r");
      fclose(fp);
    }

    if (flag == REPRIEVE)
    {
      if ((fp = fopen(dev, "w")) == (FILE *) NULL)
          bailout("can't open user's terminal", 5);
      fprintf(fp,"A line has become available you can stay :-)\n\r");
      fclose(fp);
    }

    if (flag == WARNING) {  /* process warning message */
        if ((fp = fopen(dev, "w")) == (FILE *) NULL)
            bailout("can't open user's terminal", 5);
        fprintf(fp,
            "%s: You have been online %d min.\07\n", name, idle); 
        fprintf(fp,
            "Warning, 2 min till auto-logoff...\n\r");
        fclose(fp);
        fprintf(log, "** WARNING ** %s %s (%d min online time) %s",
            name, dev+5, idle, ctime(&tvec)+3);
    }
    if (flag == LOGOFF) {   /* process log-off message */
        fprintf(log, "** LOGOFF  ** %s %s (%d min online) %s",
            name, dev + 5, idle, ctime(&tvec) + 3);
        sprintf(mbuf, "/bin/mail %s", name);
        /* open pipe to mail program for writing */
/*        if ((mprog = popen(mbuf, "w")) == (FILE *) NULL)
            bailout("can't use /bin/mail program", 6);
        fprintf(mprog, "Subject: Time limit exceeded\nLogged \
off - time limit exceeded - %s %s\ntty = %s\n",
            name, ctime(&tvec), dev + 5);
        fclose(mprog); */
    }
    if (flag == NOLOGOFF) {  /* send mail to root if can't kill */
        fprintf(log, "** LOGOFF-FAIL ** %s (pid = %d) \
%s (%d min idle time) %s",
            name, utmpp->ut_pid, dev+5, idle, ctime(&tvec)+3);
        sprintf(mbuf, "/bin/mail root");
        if ((mprog = popen(mbuf, "w")) == (FILE *) NULL)
            bailout("can't use /bin/mail program", 7);
        fprintf(mprog, "Subject: Can't logoff %s\nCan't Log off \
- %s %s\ntty = %s\n", name, name, ctime(&tvec), dev+5);
        fclose(mprog);
    }
    fclose(log);
    return(0);
}

killit(pid) /* terminate process using SIGHUP, then SIGKILL */
int pid;
{
    printf("pid : %d\n", pid);
    kill(pid, SIGHUP);  /* first send "hangup" signal */
    sleep(KWAIT);
    if (!kill(pid, 0)) {    /* SIGHUP might be ignored */
        kill(pid, SIGKILL); /* then send sure "kill" signal */
        sleep(KWAIT);
        if (!kill(pid, 0))
            return(0);  /* failure--refuses to die! */
        else
            return(1);  /* successful kill with SIGKILL */
    } else
        return(1);  /* successful kill with SIGHUP */
}

bailout(message, status) /* display error message and exit */
int     status;     /* exit status */
char    *message;   /* pointer to the error message */
{
    fprintf(stderr, "autologout: %s\n", message);
    exit(status);
}
