/*
		"@(#) autologout.c by David Dickson" 
               
                Modified by John Auckett 16/09/94

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/
#include    <fcntl.h>
#include    <stdio.h>
#include    <signal.h>
#include    <string.h>
#include    <sys/types.h>
#include    <sys/stat.h>
#include    <utmp.h>
#include    <time.h>

#define     GRACE       60  /* grace time (sec) for user reply */
#define     KWAIT       20  /* time to wait after kill (in sec.) */
#define     WARNING     1   /* a warning message */
#define     LOGOFF      2   /* a log-off message */
#define     NOLOGOFF    3   /* a log-off failure message */
#define     SEC         60  /* seconds per minute */
#define     TIMEMAX     10  /* maximum idle time (minutes) */

#define     NO_OF_LINES 5   /* number of dial-in lines installed */
#define     REPRIEVE    4   /* line became available during grace time */
#define     EXTRATIME   5   /* go into borrowed time message */
#define     DAILY_LIMIT_WARN 6
#define     LOGOFF_DAILY_LIMIT 7
#define     START_DAILY  16 /* hour at which daily limits are applicable */
#define     END_DAILY    0  /* hour at which daily limits end */
#define     DEFAULTTIME 10  /* default time limit */

struct utmp *utmpp;         /* pointer to utmp file entry */
char        *ctime();       /* returns pointer to time string */
struct utmp *getutent();    /* returns next utmp file entry */
void         setutent();
FILE        *ufp;

int check_daily_limit(char *name, int online_time, int time_allowed);

int free_lines()    /* returns the number of free dial-in lines */
{
  int line_count = 0;
  struct utmp *upp;
  
  /* count up number of ttyS devices */
  setutent();
  while ((upp = getutent()) != (struct utmp *) NULL)
    if ((upp->ut_type == USER_PROCESS) && (upp->ut_line[3] == 'S'))
      ++line_count;
/*  printf("Lines in use %d\n", line_count); */
  return(NO_OF_LINES - line_count);
}

int time_permitted(char *name)  /* returns a users time per call */
{
  FILE *fp;
  char record_name[9];
  int time_per_call;

  /* fetch a user record from dialin.users */
  if (!(fp = fopen("/usr/etc/dialin.users", "r")))
  {                   /* if the file can't be opened then    */
    fclose(fp);       /* use a default time limit            */
    return(DEFAULTTIME);
  }
  while(1)     /* continue until name found or can't read */
  {            /* anymore names from file                 */
    if (fscanf(fp, "%s %d", record_name, &time_per_call) != 2)
    {                          /* got to end without finding name */
      fclose(fp);              /* name not in the dialin.users    */
      return(DEFAULTTIME);     /* file so use default time limit  */
    }
    if (!strcmp(name, record_name))
    {   /* found the users record in the file */
      fclose(fp);
      return(time_per_call);
    }
  }
}

main()
{ 
  int count=0;
  
    while ((utmpp = getutent()) != (struct utmp *) NULL)
    {
      if (utmpp->ut_type == USER_PROCESS)
        if (fork())     /* parent process continues thru utmp entries */
          sleep(2);     /* sleep here so children don't compete for access to user_times file*/
        else
        {
          check_idle(); /* child checks utmp entries */
        }
     }
}

int check_daily_limit(char *name, int online_time, int time_allowed)
{
  /* returns zero if passed the check 1 if failed (ie over limit) */

  time_t local_time;
  struct tm *tm_local; 
  int mins_since_midnight;
  int tl;
  int hr;

  struct user_time_record
  {
    char name[9];
    int  time;
  };
  int num_read;

  struct user_time_record user_time;

  local_time = time((time_t *)0);
  tm_local = localtime(&local_time);

  /* check if we are in the time for logging daily limits */
/* make it simple for now */


  hr = tm_local->tm_hour;
  if (hr <= START_DAILY)
    return(0);

/*  if (DAILY_END < DAILY_START) 
  { 
    if (hr > DAILY_END || hr < DAILY_START)
      return(0); 
  }
  else 
    if (!((hr < DAILY_END) && (hr >= DAILY_START)))
      return(0);
*/

  /* open the users daily time file */
  if ((tl = open("/usr/etc/user_times", O_RDWR)) == -1)
  {
    printf("unable to open user_times file\n\r");
    tl = creat("/usr/etc/user_times", 0644);
    close(tl);
    if ((tl = open("/usr/etc/user_times", O_RDWR)) == -1) 
      return(0); /* problem opening it so just bail */
  }

  /* locate the users name in the file */
 
  num_read = read(tl, &user_time, sizeof(user_time)); 
  while(num_read == sizeof(user_time)) /* continue until name found or can't read */
  {            
    if (!strcmp(name, user_time.name))
    {   /* found the users record in the file */
      /* update the record first */

      lseek(tl, -sizeof(user_time), 1);
      if (online_time < 5)
        user_time.time += online_time;
      else
        user_time.time += 5;  
      write(tl, &user_time, sizeof(user_time));
      close(tl);

      /* now do the checking */
      if (user_time.time > (time_allowed * 2))
        return(1); /* failed the check */
      else
        return(0);
    }
    num_read = read(tl, &user_time, sizeof(user_time));
  }
  /* got to end without finding it so create it */
  /* calculate the time since midnight */
  /* if online time > time since midnight */
  /* then time is only the time since midnight */
  
  mins_since_midnight = (tm_local->tm_hour * 60) + tm_local->tm_min; 
  printf("New entry %s %d\n\r", name, mins_since_midnight);
  if (online_time > mins_since_midnight)
    user_time.time = mins_since_midnight;
  else
    user_time.time = online_time;
  strcpy(user_time.name, name);
  write(tl, &user_time, sizeof(user_time));
  close(tl);
  return(0);
}

check_idle()    /* select utmp entries older than TIMEMAX */
{
    int lf;
    char        dev[24], name[12];
    struct stat status, *pstat;
    time_t      online, pres_time, start, time();
    int         time_allowed;
    int         user_pid;
    struct utmp *newutmp;
    uid_t       user_id;
    int daily_time;
    char tmp_buf[80];


    pstat = &status;    /* point to status structure */

    if (utmpp->ut_line[3] != 'S')
    /*  && (utmpp->ut_line[3] != 'p'))  */
                 /* if not a dial in line or a telnet*/
        exit(0);                        /* skip the utmp entry   */

    strcpy(dev, "/dev/");           /* add "/dev/" directory prefix */
    strcat(dev, utmpp->ut_line);    /* append basename of port */
    if (stat(dev, pstat))           /* if can't get status for port */
        bailout("can't get status of user's terminal", 1);
    pres_time = time((time_t *)0);  /* get current time */

    online = (pres_time - utmpp->ut_time) /SEC;
 
    strncpy(name, utmpp->ut_user, 8);   /* else get user name */
    name[8] = '\0';          /* null terminate user name string */
    
    /* calling free_lines() corrupts the utmpp pointer */ 
    user_pid = utmpp->ut_pid;

    time_allowed = time_permitted(name);
/* check the daily limit here */
    if (check_daily_limit(name, online,time_allowed))
    { /* returns true if failed the daily limit check (ie over limit ) */
      /* just log it for now */
      if ((lf = open("/usr/adm/dailylimitlog", O_WRONLY)) != -1)
      {
        lseek(lf, 0L, 2);
        sprintf(tmp_buf, "*** %s exceeded daily limit ***\n", name);
        write(lf, tmp_buf, strlen(tmp_buf));
        close(lf);
        system("date >> /usr/adm/dailylimitlog");
      }
      mesg(DAILY_LIMIT_WARN, name, dev, (int)online); /* send warning to user */
      if (stat(dev, pstat))
          bailout("can't get status of user's terminal", 2);
      
      start = pstat->st_atime;    /* start time for countdown */
    
      if (time_allowed > 25)    /* only allow borrowed time for */
        if (free_lines() > 0)   /* non-guest accounts           */
        {
            mesg(EXTRATIME, name, dev, (int)online);
            exit(0);
        }

      user_id = pstat->st_uid;  /* remember the user id */

    /* sleep for 2 mins */
      sleep(GRACE*2);
    
      if (stat(dev, pstat))
          bailout("can't get status of user's terminal", 3);
      if (pstat->st_uid != user_id)
          bailout("autologout left over from dropped carrier", 3);

      if (time_allowed > 25)    /* only all borowed time for */
        if (free_lines() > 0)     /* non-guests                */
        {
            mesg(REPRIEVE, name, dev, (int)online);
            exit(0);
        }
        /* user has run out of time */
/* won't kill them until we have logs to make sure its working */
      mesg(LOGOFF_DAILY_LIMIT, name, dev, (int)online);
#if 1
      if (!killit(user_pid))
        mesg(NOLOGOFF, name, dev, (int)online); /* couldn't kill */
      else
        mesg(LOGOFF_DAILY_LIMIT, name, dev, (int)online); /* okay */
      exit(0);
#endif
    }

/* the rest of this function is for call limit */
    if (online < time_allowed)
      exit(0);

    mesg(WARNING, name, dev, (int)online); /* send warning to user */
    if (stat(dev, pstat))
        bailout("can't get status of user's terminal", 2);
    start = pstat->st_atime;    /* start time for countdown */
    
    if (time_allowed > 25)    /* only allow borrowed time for */
      if (free_lines() > 0)   /* non-guest accounts           */
      {
          mesg(EXTRATIME, name, dev, (int)online);
          exit(0);
      }

user_id = pstat->st_uid;  /* remember the user id */

/* sleep for 2 mins */
    sleep(GRACE*2);
    
    if (stat(dev, pstat))
        bailout("can't get status of user's terminal", 3);
    if (pstat->st_uid != user_id)
        bailout("autologout left over from dropped carrier", 3);

    if (time_allowed > 25)    /* only all borowed time for */
      if (free_lines() > 0)     /* non-guests                */
      {
          mesg(REPRIEVE, name, dev, (int)online);
          exit(0);
      }
      /* user has run out of time */
    if (!killit(user_pid))
      mesg(NOLOGOFF, name, dev, (int)online); /* couldn't kill */
    else
      mesg(LOGOFF, name, dev, (int)online); /* okay */
    exit(0);
}

mesg(flag, name, dev, idle) /* mail to user and log message */
int     flag, idle;     /* flag indicates message type */
char    *name, *dev;
{
    char    mbuf[256];  /* message buffer */
    time_t  tvec;
    FILE    *fopen(), *fp, *log, *mprog;

    time(&tvec);    /* store time in tvec */
    if ((log = fopen("/usr/adm/logoutlog", "a")) == (FILE *) NULL)
        bailout("can't open log file", 4);

    if (flag == EXTRATIME)
    {
      if ((fp = fopen(dev, "w")) == (FILE *) NULL)
          bailout("can't open user's terminal", 5);
      fprintf(fp, "But since a line is free you can stay...\n\r");
      fclose(fp);
    }

    if (flag == DAILY_LIMIT_WARN)
    {
      if ((fp = fopen(dev, "w")) == (FILE *) NULL)
        bailout("can't open user's terminal", 5);
      fprintf(fp, "***Warning***, Daily time limit exceeded, 2 minutes till autologoff...\n\r");
      fclose(fp); 
      fprintf(log, "** DAILY LIMIT WARNING ** %s %s (%d min online time) %s",
              name, dev+5, idle, ctime(&tvec)+3);

    } 

    if (flag == LOGOFF_DAILY_LIMIT)    /* process log-off message */
    {
        fprintf(log, "** LOGOFF DAILY LIMIT ** %s %s (%d min online) %s",
            name, dev + 5, idle, ctime(&tvec) + 3);
        sprintf(mbuf, "/bin/mail %s", name);
    }
   
    if (flag == REPRIEVE)
    {
      if ((fp = fopen(dev, "w")) == (FILE *) NULL)
          bailout("can't open user's terminal", 5);
      fprintf(fp,"A line has become available you can stay :-)\n\r");
      fclose(fp);
    }

    if (flag == WARNING) {  /* process warning message */
        if ((fp = fopen(dev, "w")) == (FILE *) NULL)
            bailout("can't open user's terminal", 5);
        fprintf(fp,
            "%s: You have been online %d min.\07\n", name, idle); 
        fprintf(fp,
            "Warning, 2 min till auto-logoff...\n\r");
        fclose(fp);
        fprintf(log, "** WARNING ** %s %s (%d min online time) %s",
            name, dev+5, idle, ctime(&tvec)+3);
    }
    if (flag == LOGOFF) {   /* process log-off message */
        fprintf(log, "** LOGOFF  ** %s %s (%d min online) %s",
            name, dev + 5, idle, ctime(&tvec) + 3);
        sprintf(mbuf, "/bin/mail %s", name);
        /* open pipe to mail program for writing */
/*        if ((mprog = popen(mbuf, "w")) == (FILE *) NULL)
            bailout("can't use /bin/mail program", 6);
        fprintf(mprog, "Subject: Time limit exceeded\nLogged \
off - time limit exceeded - %s %s\ntty = %s\n",
            name, ctime(&tvec), dev + 5);
        fclose(mprog); */
    }
    if (flag == NOLOGOFF) {  /* send mail to root if can't kill */
        fprintf(log, "** LOGOFF-FAIL ** %s (pid = %d) \
%s (%d min idle time) %s",
            name, utmpp->ut_pid, dev+5, idle, ctime(&tvec)+3);
        sprintf(mbuf, "/bin/mail root");
        if ((mprog = popen(mbuf, "w")) == (FILE *) NULL)
            bailout("can't use /bin/mail program", 7);
        fprintf(mprog, "Subject: Can't logoff %s\nCan't Log off \
- %s %s\ntty = %s\n", name, name, ctime(&tvec), dev+5);
        fclose(mprog);
    }
    fclose(log);
    return(0);
}

killit(pid) /* terminate process using SIGHUP, then SIGKILL */
int pid;
{
    printf("pid : %d\n", pid);
    kill(pid, SIGHUP);  /* first send "hangup" signal */
    sleep(KWAIT);
    if (!kill(pid, 0)) {    /* SIGHUP might be ignored */
        kill(pid, SIGKILL); /* then send sure "kill" signal */
        sleep(KWAIT);
        if (!kill(pid, 0))
            return(0);  /* failure--refuses to die! */
        else
            return(1);  /* successful kill with SIGKILL */
    } else
        return(1);  /* successful kill with SIGHUP */
}

bailout(message, status) /* display error message and exit */
int     status;     /* exit status */
char    *message;   /* pointer to the error message */
{
    fprintf(stderr, "autologout: %s\n", message);
    exit(status);
}
